//
// Created by vlad on 23.03.23.
//

#include "MeshLoader.h"
#include <vector>
#include <glm/vec3.hpp>
#include <fstream>
#include <iostream>
#include <glm/vec2.hpp>
#include <sstream>


enum ParserState {
    IDLE,
    COMMENT,
    READ_STR
};

struct MeshPoint {
    glm::vec3 point;
    glm::vec3 normal;
};

Mesh *MeshLoader::loadMesh(const std::string &filename) {
    std::vector<glm::vec3> points;
    std::vector<glm::vec3> normals;
    std::vector<MeshPoint> mesh_points;


    std::ifstream file(filename + ".obj", std::ios::in);

    ParserState state = IDLE;
    int line = 1;

    char tmp_buff[128];
    int tmp_buff_index = 0;
    while (true) {
        char in;

        file.read(&in, 1);

        if (file.eof()) {
            break;
        }

        if (tmp_buff_index >= 128) {
            std::cerr << "[MeshLoader] Error: tmp buffer overflow"
                         " (line " << line << " file: " << filename << ")" << std::endl;
            return nullptr;
        }

        if (in == '\n') {
            line++;
        }

        switch (state) {
            case IDLE:
                tmp_buff[0] = in;
                tmp_buff_index = 1;

                // нет состояния, возможно после пустой строки
                if (in == '#' || in == 'o' || in == 's') {
                    // коммент если это информация о имени объекта или о сглаживании
                    state = COMMENT;
                } else if (in == 'v' || in == 'f') {
                    state = READ_STR;
                }
                break;

            case COMMENT:
                // строка коммента, читаем до конца линии
                if (in == '\n') {
                    state = IDLE;
                }
                break;

            case READ_STR:
                if (in == '\n') {
                    state = IDLE;
                    // парсим строку
                    tmp_buff[tmp_buff_index] = 0; // делаем null-terminator
                    tmp_buff_index = 0;
                    std::string source(tmp_buff);

                    if (source.find("v ") == 0) {
                        // вершина
                        char* end = nullptr;

                        glm::vec3 tmp_point(0, 0, 0);
                        tmp_point.x = std::strtof(tmp_buff + 2, &end);
                        tmp_point.y = std::strtof(end, &end);
                        tmp_point.z = std::strtof(end, &end);
                        points.push_back(tmp_point);
                    } else if (source.find("vn ") == 0) {
                        // нормаль
                        char* end = nullptr;

                        glm::vec3 tmp_point(0, 0, 0);
                        tmp_point.x = std::strtof(tmp_buff + 3, &end);
                        tmp_point.y = std::strtof(end, &end);
                        tmp_point.z = std::strtof(end, &end);
                        normals.push_back(tmp_point);
                    } else if (source.find("f ") == 0) {
                        // фрагмент, он же треугольник
                        std::istringstream source_stream(tmp_buff + 2);
                        std::string point;
                        while (getline(source_stream, point, ' ')) {
                            std::istringstream point_stream(point);
                            std::string vertex, texture, normal;
                            getline(point_stream, vertex, '/');
                            getline(point_stream, texture, '/');
                            getline(point_stream, normal, '\0'); // до конца строки

                            MeshPoint p{};
                            // сначала точка
                            int tmp = std::atoi(vertex.c_str());
                            if (tmp > 0) {
                                tmp -= 1;
                                p.point = points[tmp];
                            }

                            tmp = std::atoi(normal.c_str());
                            if (tmp > 0) {
                                tmp -= 1;
                                p.normal = normals[tmp];
                            }

                            mesh_points.push_back(p);
                        }
                    }
                } else {
                    tmp_buff[tmp_buff_index++] = in;
                }
                break;
        }
    }

    // тут имеем готовый массив вершин
    // размер меша = кол-во вершин * (3 + 3)
    const int vertices = (int)mesh_points.size();
    auto* out_mesh = new float[vertices * (3 + 3)];

    int curr_index = 0;
    for (const auto& p : mesh_points) {
        // точка
        out_mesh[curr_index++] = p.point.x;
        out_mesh[curr_index++] = p.point.y;
        out_mesh[curr_index++] = p.point.z;

        // нормаль
        out_mesh[curr_index++] = p.normal.x;
        out_mesh[curr_index++] = p.normal.y;
        out_mesh[curr_index++] = p.normal.z;
    }

    const int attrs[] = {3, 3, 0};
    Mesh* m = new Mesh(out_mesh, vertices, attrs);

    delete[] out_mesh;
    return m;
}
